/*
 * Copyright (c) 2017-2022, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <linux/err.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/vfio.h>

#include "nvstatus.h"
#include "nv-linux.h"
#include "nv-vgpu-vfio-interface.h"
#include "vgpu-devices.h"

#if defined(NV_VGPU_KVM_BUILD) && defined(NV_MDEV_PRESENT)

#if defined(NV_MDEV_PARENT_STRUCT_PRESENT)
struct mdev_driver nv_vfio_mdev_driver;
#endif

/* Common code used by both modern and older mdev */

#if defined(DEBUG)
static ssize_t
migration_data_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    ssize_t     ret;
    vgpu_dev_t *vgpu_dev = NULL;

    down(&vgpu_devices.vgpu_dev_list_lock);
#if defined(NV_USE_MODERN_MDEV)
    vgpu_dev = dev_get_drvdata(dev);
#else
    vgpu_dev = mdev_get_drvdata(NV_GET_MDEV_FROM_DEV(dev));
#endif
    if (vgpu_dev == NULL)
    {
        up(&vgpu_devices.vgpu_dev_list_lock);
        return -ENXIO;
    }
    down(&vgpu_dev->ops_lock);
    up(&vgpu_devices.vgpu_dev_list_lock);

#if defined(NV_VFIO_DEVICE_MIG_STATE_PRESENT) || defined(NV_VFIO_DEVICE_MIGRATION_HAS_START_PFN)
    if (vgpu_dev->migration_info.migration_state == NV_VFIO_DEVICE_STATE_NONE)
    {
        ret = sprintf(buf,
                "Bytes Transferred:\n\tPre_copy: %lld (%lld MB)\n\tStop_copy: %lld (%lld MB)\nDirty Pages: %d\n",
                vgpu_dev->migration_info.bytes_transferred_precopy,
                vgpu_dev->migration_info.bytes_transferred_precopy >> 20,
                vgpu_dev->migration_info.bytes_transferred_stopcopy,
                vgpu_dev->migration_info.bytes_transferred_stopcopy >> 20,
                vgpu_dev->migration_info.dirty_pfn_count);
    }
    else
#endif
    {
        ret = sprintf(buf, "Not Migrated\n");
    }

    up(&vgpu_dev->ops_lock);

    return ret;
}
static DEVICE_ATTR_RO(migration_data);
#endif

static DEVICE_ATTR_RW(vgpu_params);

static struct attribute *vgpu_dev_attrs[] = {
    &dev_attr_vgpu_params.attr,
#if defined(DEBUG)
    &dev_attr_migration_data.attr,
#endif
    NULL,
};

static const struct attribute_group vgpu_dev_group = {
    .name  = "nvidia",
    .attrs = vgpu_dev_attrs,
};

static const struct attribute_group *vgpu_dev_groups[] = {
    &vgpu_dev_group,
    NULL,
};

static void nv_init_vgpu_dev(vgpu_dev_t *vgpu_dev, struct mdev_device *mdev)
{
    vgpu_dev->mdev = mdev;
    vgpu_dev->dev = NV_GET_MDEV_DEV(mdev);
}

int nv_vgpu_vfio_mdev_destroy(vgpu_dev_t *vgpu_dev)
{
    struct mdev_device *mdev = NULL;
    struct device *dev = NULL;
    int ret = 0;

    mdev = vgpu_dev->mdev;
    dev  = NV_GET_MDEV_DEV(mdev);

    ret = nv_vgpu_vfio_destroy(vgpu_dev, dev);
    if (ret)
        return ret;

    return 0;
}

#if defined(NV_MDEV_GET_TYPE_GROUP_ID_PRESENT) || defined(NV_MDEV_PARENT_STRUCT_PRESENT)
int nv_vgpu_vfio_mdev_create(vgpu_dev_t *vgpu_dev)
{
    struct kobject *kobj = NULL;
#else
int nv_vgpu_vfio_mdev_create(struct kobject *kobj, vgpu_dev_t *vgpu_dev)
{
#endif
    struct pci_dev *pdev;
    void *mtype = NULL;
    NvU32 vgpu_type_id = 0;
    int ret = 0;

    if (!vgpu_dev)
    {
        NV_VGPU_LOG(VGPU_ERR, "No vGPU device found\n");
        return -EINVAL;
    }

    pdev = to_pci_dev(NV_GET_MDEV_PARENT(vgpu_dev->mdev));
    if (!pdev)
        return -EINVAL;

#if defined(NV_MDEV_GET_TYPE_GROUP_ID_PRESENT) || defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    mtype = vgpu_dev->mdev->type;
#endif

    if (nv_get_vgpu_type_id(kobj, NV_GET_MDEV_PARENT(vgpu_dev->mdev), &vgpu_type_id, mtype)
        != NV_OK)
    {
        NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "vGPU creation failed, %d\n", ret);
        return -EINVAL;
    }

    ret = nv_vgpu_vfio_create(vgpu_dev, pdev, vgpu_type_id);
    if (ret)
        return ret;

    if (pdev->is_virtfn)
    {
#if defined(NV_MDEV_SET_IOMMU_DEVICE_PRESENT)

#if defined(NV_MDEV_GET_TYPE_GROUP_ID_PRESENT) || defined(NV_MDEV_PARENT_STRUCT_PRESENT)
        mdev_set_iommu_device(vgpu_dev->mdev, NV_GET_MDEV_PARENT(vgpu_dev->mdev));
#else
        ret = mdev_set_iommu_device(NV_GET_MDEV_DEV(vgpu_dev->mdev),
                                    NV_GET_MDEV_PARENT(vgpu_dev->mdev));
#endif
        if (ret != 0)
        {
            NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "Failed to set IOMMU device. ret: %d \n", ret);
            nv_vgpu_vfio_mdev_destroy(vgpu_dev);
        }
#endif
    }

    return 0;
}

NV_STATUS nv_get_vgpu_type_id(struct kobject *kobj, struct device *dev,
                                     NvU32 *vgpu_type_id, void *mtype)
{
#if defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    vgpu_type_t *vgpu_type = NULL;

    if (!mtype)
        return NV_ERR_INVALID_ARGUMENT;

    vgpu_type = container_of((mtype), vgpu_type_t, mtype);
    *vgpu_type_id = vgpu_type->vgpu_type_id;

#elif defined(NV_MDEV_GET_TYPE_GROUP_ID_PRESENT)
    struct pci_dev *pdev = to_pci_dev(dev);
    phys_dev_t *phys_dev = NULL;

    if (mtype == NULL) {
        return NV_ERR_INVALID_ARGUMENT;
    }

    down(&phys_devices.phys_dev_list_lock);
    if ((get_phys_dev(pdev, &phys_dev) != NV_OK) || (phys_dev == NULL))
    {
       up(&phys_devices.phys_dev_list_lock);
       return NV_ERR_INVALID_ARGUMENT;
    }
    up(&phys_devices.phys_dev_list_lock);

    down(&phys_dev->ops_lock);
    if (mtype_get_type_group_id(mtype) >= phys_dev->num_vgpu_types)
    {
        up(&phys_dev->ops_lock);
        return NV_ERR_INVALID_ARGUMENT;
    }

    *vgpu_type_id = phys_dev->vgpu_type_ids[mtype_get_type_group_id(mtype)];
    up(&phys_dev->ops_lock);

#else
    char *strp = NULL;
    const char *key = dev_driver_string(dev);

    if (kobj == NULL)
        return NV_ERR_INVALID_ARGUMENT;

    strp = strstr(kobj->name, key);
    if (strp == NULL)
        return NV_ERR_INVALID_ARGUMENT;

    strp = strp + strlen(key) + 1;
    *vgpu_type_id = strtoul(strp, NULL, 0);
#endif
    return NV_OK;
}

int nv_mdev_register(phys_dev_t *phys_dev, struct pci_dev *pdev)
{
#if defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    return mdev_register_parent(&phys_dev->vgpu_parent, &pdev->dev,
                                &nv_vfio_mdev_driver, phys_dev->vgpu_mdev_types,
                                phys_dev->num_vgpu_types);
#else
    return mdev_register_device(&pdev->dev, phys_dev->vgpu_fops);
#endif
}

void nv_mdev_unregister(phys_dev_t *phys_dev, struct pci_dev *pdev)
{
#if defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    mdev_unregister_parent(&phys_dev->vgpu_parent);
#else
    mdev_unregister_device(&pdev->dev);
#endif
}

void nv_vgpu_free_mdev_fops(struct phys_dev_s *phys_dev)
{
#if !defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    if (!phys_dev->vgpu_fops)
        return;

#if defined(NV_MDEV_DRIVER_HAS_SUPPORTED_TYPE_GROUPS)
    NV_KFREE(phys_dev->vgpu_fops, sizeof(struct mdev_driver));
#elif defined(NV_MDEV_PARENT_OPS_STRUCT_PRESENT)
    NV_KFREE(phys_dev->vgpu_fops, sizeof(struct mdev_parent_ops));
#else
    NV_KFREE(phys_dev->vgpu_fops, sizeof(struct parent_ops));
#endif

#endif /* !NV_MDEV_PARENT_STRUCT_PRESENT */
}

#if defined(NV_USE_MODERN_MDEV)

/* Modern mdev codepath for kernel > v5.14 */

extern rm_vgpu_vfio_ops_t rm_vgpu_vfio_ops;

static vgpu_dev_t *nv_get_vgpu_from_vdev(struct vfio_device *core_vdev)
{
    vgpu_dev_t *vgpu_dev= NULL;

    if (!core_vdev || !core_vdev->dev)
    {
        NV_VGPU_LOG(VGPU_ERR, "No vfio device found\n");
        return NULL;
    }

    vgpu_dev = dev_get_drvdata(core_vdev->dev);
    if (!vgpu_dev)
        return NULL;

    return vgpu_dev;
}

static int nv_vfio_mdev_open(struct vfio_device *core_vdev)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    int ret;

    ret = nv_vgpu_vfio_open(vgpu_dev);
    if (ret)
        return ret;

    return ret;
}

static void nv_vfio_mdev_release(struct vfio_device *core_vdev)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    nv_vgpu_vfio_close(vgpu_dev);
}

static long nv_vfio_mdev_unlocked_ioctl(struct vfio_device *core_vdev,
                     unsigned int cmd, unsigned long arg)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    return nv_vgpu_vfio_ioctl(vgpu_dev, cmd, arg);
}

static ssize_t nv_vfio_mdev_read(struct vfio_device *core_vdev, char __user *buf,
                  size_t count, loff_t *ppos)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    return nv_vgpu_vfio_read(vgpu_dev, buf, count, ppos);
}

static ssize_t nv_vfio_mdev_write(struct vfio_device *core_vdev,
                   const char __user *buf, size_t count,
                   loff_t *ppos)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    return nv_vgpu_vfio_write(vgpu_dev, buf, count, ppos);
}

static int nv_vfio_mdev_mmap(struct vfio_device *core_vdev,
              struct vm_area_struct *vma)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    return nv_vgpu_vfio_mmap(vgpu_dev, vma);
}

static void nv_vfio_mdev_request(struct vfio_device *core_vdev, unsigned int count)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    if (!vgpu_dev)
        return;

    if (count == 0)
        NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev,
                    "No mdev vendor driver request callback support, blocked until released by user\n");
}

#if defined(NV_VFIO_DEVICE_MIG_STATE_PRESENT)
static int nv_vfio_mdev_get_device_state(struct vfio_device *core_vdev,
                                    enum vfio_device_mig_state *curr_state)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    if (!vgpu_dev)
        return -EINVAL;

    *curr_state = vgpu_dev->migration_info.vfio_state;

    return 0;
}

static struct file *
nv_vfio_mdev_set_device_state(struct vfio_device *core_vdev,
                         enum vfio_device_mig_state new_state)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    struct file *fp = NULL;

    if (!vgpu_dev)
        return ERR_PTR(-EINVAL);

    fp = nv_vgpu_set_device_state(core_vdev, vgpu_dev, new_state);

    return fp;
}
#endif

#if defined(NV_VFIO_DEVICE_OPS_HAS_DMA_UNMAP)
static void nv_vfio_mdev_dma_unmap(struct vfio_device *core_vdev,
                                   u64 iova, u64 size)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);
    if (!vgpu_dev)
        return;

    nv_vgpu_vfio_dma_unmap(vgpu_dev, iova, size);
}
#endif

static int nv_vfio_mdev_get_migration_data_size(struct vfio_device *core_vdev, unsigned long *size)
{
    int ret = 0;
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_vdev(core_vdev);

    if (!vgpu_dev)
        return -EINVAL;

    ret = nv_vgpu_vfio_get_migration_data_size(vgpu_dev, size);

    return ret;
}

#if defined(NV_VFIO_MIGRATION_OPS_PRESENT)
static const struct vfio_migration_ops nv_vfio_mdev_migration_ops = {
    .migration_set_state = nv_vfio_mdev_set_device_state,
    .migration_get_state = nv_vfio_mdev_get_device_state,
#if defined(NV_VFIO_MIGRATION_OPS_HAS_MIGRATION_GET_DATA_SIZE)
    .migration_get_data_size = nv_vfio_mdev_get_migration_data_size,
#endif
};
#endif

#if defined(NV_VFIO_LOG_OPS_PRESENT)
static int nv_vfio_mdev_start_page_tracker(struct vfio_device *vdev,
                                           struct rb_root_cached *ranges,
                                           u32 nnodes, u64 *page_size)
{
    return 0;
}

static int nv_vfio_mdev_stop_page_tracker(struct vfio_device *vdev)
{
    return 0;
}

static int nv_vfio_mdev_tracker_read_and_clear(struct vfio_device *vdev,
                                               unsigned long iova,
                                               unsigned long length,
                                               struct iova_bitmap *dirty)
{
    vgpu_dev_t *vgpu_dev = NULL;

    if (!vdev)
        return -EINVAL;

    vgpu_dev = nv_get_vgpu_from_vdev(vdev);
    return nv_vfio_tracker_read_and_clear(vgpu_dev, iova, length, dirty);
}

static const struct vfio_log_ops nv_vfio_mdev_log_ops = {
    .log_start = nv_vfio_mdev_start_page_tracker,
    .log_stop = nv_vfio_mdev_stop_page_tracker,
    .log_read_and_clear = nv_vfio_mdev_tracker_read_and_clear,
};
#endif /* NV_VFIO_LOG_OPS_PRESENT */

#if defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
static int nv_vfio_mdev_init_dev(struct vfio_device *core_vdev)
{
    vgpu_dev_t *vgpu_dev = NULL;
    struct mdev_device *mdev = NULL;
    int ret = 0;

    if (!core_vdev)
    {
        NV_VGPU_LOG(VGPU_ERR, "No vfio device found\n");
        return -EINVAL;
    }

    mdev = NV_GET_MDEV_FROM_DEV(core_vdev->dev);
    if (!mdev)
    {
        NV_VGPU_LOG(VGPU_ERR, "No mdev device found\n");
        return -ENODEV;
    }

    vgpu_dev = container_of(core_vdev, vgpu_dev_t, vfio_vdev);
    if (!vgpu_dev)
    {
        NV_VGPU_LOG(VGPU_ERR, "No vGPU device found\n");
        return -ENODEV;
    }

    nv_init_vgpu_dev(vgpu_dev, mdev);

    ret = nv_vgpu_vfio_mdev_create(vgpu_dev);
    if (ret)
    {
        NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "Failed to create mdev device\n");
        return ret;
    }

    return ret;
}

static void nv_vfio_mdev_release_dev(struct vfio_device *core_vdev)
{
    vgpu_dev_t *vgpu_dev = NULL;
    int ret = 0;

    if (!core_vdev)
        return;

    vgpu_dev = container_of(core_vdev, vgpu_dev_t, vfio_vdev);
    if (!vgpu_dev)
        return;

    ret = nv_vgpu_vfio_mdev_destroy(vgpu_dev);
    if (ret)
        NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "Failed to destroy vGPU device, ret: %d\n", ret);

#if defined(NV_VFIO_FREE_DEVICE_PRESENT)
    vfio_free_device(vgpu_dev->vdev);
#endif
}
#endif

static const struct vfio_device_ops nv_vfio_mdev_dev_ops = {
    .name           = "nvidia-vgpu-vfio",
#if defined(NV_MDEV_PARENT_OPS_HAS_OPEN_DEVICE) || defined(NV_MDEV_DRIVER_HAS_SUPPORTED_TYPE_GROUPS) || defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    .open_device    = nv_vfio_mdev_open,
    .close_device   = nv_vfio_mdev_release,
#else
    .open           = nv_vfio_mdev_open,
    .release        = nv_vfio_mdev_release,
#endif
    .ioctl          = nv_vfio_mdev_unlocked_ioctl,
    .read           = nv_vfio_mdev_read,
    .write          = nv_vfio_mdev_write,
    .mmap           = nv_vfio_mdev_mmap,
    .request        = nv_vfio_mdev_request,
#if defined(NV_VFIO_DEVICE_OPS_HAS_DMA_UNMAP)
    .dma_unmap      = nv_vfio_mdev_dma_unmap,
#endif
#if defined(NV_VFIO_DEVICE_MIG_STATE_PRESENT) && !defined(NV_VFIO_MIGRATION_OPS_PRESENT)
    .migration_set_state = nv_vfio_mdev_set_device_state,
    .migration_get_state = nv_vfio_mdev_get_device_state,
#endif
#if defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    .init           = nv_vfio_mdev_init_dev,
    .release        = nv_vfio_mdev_release_dev,
#endif
#if defined(NV_VFIO_DEVICE_OPS_HAS_BIND_IOMMUFD)
    .bind_iommufd   = vfio_iommufd_emulated_bind,
    .unbind_iommufd = vfio_iommufd_emulated_unbind,
    .attach_ioas    = vfio_iommufd_emulated_attach_ioas,
#endif
};

void nv_vfio_mdev_set_mig_ops(struct vfio_device *core_vdev, struct pci_dev *pdev)
{
#if defined(NV_VFIO_PRECOPY_INFO_PRESENT)
    // Precopy is only supported on VF
    if (pdev->is_virtfn)
        core_vdev->migration_flags = VFIO_MIGRATION_STOP_COPY | VFIO_MIGRATION_PRE_COPY;
    else
        core_vdev->migration_flags = VFIO_MIGRATION_STOP_COPY;
#elif defined(NV_VFIO_DEVICE_MIG_STATE_PRESENT)
    core_vdev->migration_flags = VFIO_MIGRATION_STOP_COPY;
#endif
#if defined(NV_VFIO_MIGRATION_OPS_PRESENT)
    core_vdev->mig_ops = &nv_vfio_mdev_migration_ops;
#endif
#if defined(NV_VFIO_LOG_OPS_PRESENT)
    core_vdev->log_ops = &nv_vfio_mdev_log_ops;
#endif
}

static int nv_vfio_mdev_probe(struct mdev_device *mdev)
{
    struct vfio_device *vdev;
    struct pci_dev *pdev;
    vgpu_dev_t *vgpu_dev = NULL;
    int ret = 0;

    if (!mdev)
        return -EINVAL;

#if defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    vgpu_dev = vfio_alloc_device(vgpu_dev_s, vfio_vdev, &mdev->dev, &nv_vfio_mdev_dev_ops);
    if (IS_ERR(vgpu_dev))
    {
        NV_VGPU_LOG(VGPU_ERR, "Failed to allocate vGPU device\n");
        return -ENOMEM;
    }

    vdev = &vgpu_dev->vfio_vdev;
#else
    ret = nv_alloc_vgpu_dev(&vgpu_dev);
    if (ret)
        return ret;

    nv_init_vgpu_dev(vgpu_dev, mdev);

    vdev = kzalloc(sizeof(*vdev), GFP_KERNEL);
    if (!vdev)
        return -ENOMEM;

    vfio_init_group_dev(vdev, &mdev->dev, &nv_vfio_mdev_dev_ops);
#endif

    pdev = to_pci_dev(NV_GET_MDEV_PARENT(vgpu_dev->mdev));
    nv_vfio_mdev_set_mig_ops(vdev, pdev);

#if defined(NV_VFIO_REGISTER_EMULATED_IOMMU_DEV_PRESENT)
    ret = vfio_register_emulated_iommu_dev(vdev);
#else
    ret = vfio_register_group_dev(vdev);
#endif
    if (ret)
        goto out_err;

#if !defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    ret = nv_vgpu_vfio_mdev_create(vgpu_dev);
    if (ret)
    {
        vfio_unregister_group_dev(vdev);
        goto out_err;
    }
#endif

    dev_set_drvdata(&mdev->dev, vgpu_dev);
    vgpu_dev->vdev = vdev;
    return 0;

out_err:
    NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "Failed to probe mdev device, ret: %d\n", ret);

#if defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    vfio_put_device(vdev);
#else /* NV_VFIO_ALLOC_DEVICE_PRESENT */
#if defined(NV_VFIO_UNINIT_GROUP_DEV_PRESENT)
    vfio_uninit_group_dev(vdev);
#endif /* NV_VFIO_UNINIT_GROUP_DEV_PRESENT */
    kfree(vdev);
    nv_free_vgpu_dev(vgpu_dev);
#endif /* NV_VFIO_ALLOC_DEVICE_PRESENT */
    return ret;
}

static void nv_vfio_mdev_remove(struct mdev_device *mdev)
{
#if !defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    int ret;
#endif
    struct vfio_device *vdev = NULL;
    vgpu_dev_t *vgpu_dev = NULL;

    if (!mdev)
        return;

    vgpu_dev = dev_get_drvdata(&mdev->dev);
    if (!vgpu_dev)
        return;

    vdev = vgpu_dev->vdev;
    if (!vdev)
    {
        NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "Failed to get device driver data\n");
        return;
    }

    vfio_unregister_group_dev(vdev);

#if !defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    ret = nv_vgpu_vfio_mdev_destroy(vgpu_dev);
    if (ret)
        NV_VGPU_DEV_LOG(VGPU_ERR, vgpu_dev, "Failed to destroy vGPU device, ret: %d\n", ret);

#endif

    dev_set_drvdata(&mdev->dev, NULL);

#if defined(NV_VFIO_ALLOC_DEVICE_PRESENT)
    vfio_put_device(vdev);
#else /* NV_VFIO_ALLOC_DEVICE_PRESENT */
#if defined(NV_VFIO_UNINIT_GROUP_DEV_PRESENT)
    vfio_uninit_group_dev(vdev);
#endif /* NV_VFIO_UNINIT_GROUP_DEV_PRESENT */
    kfree(vdev);
    nv_free_vgpu_dev(vgpu_dev);
#endif /* NV_VFIO_ALLOC_DEVICE_PRESENT */
}

#if defined(NV_MDEV_PARENT_STRUCT_PRESENT)
static unsigned int nv_vfio_mdev_get_available(struct mdev_type *mtype)
{
    NvU32 vgpu_type_id = 0;
    struct device *dev = NV_MTYPE_GET_PARENT_DEV(mtype);
    struct pci_dev *pdev = to_pci_dev(dev);
    char buf[NV_VGPU_STRING_BUFFER_SIZE];
    unsigned int available_instances;
    NV_STATUS status;
    int ret;

    status = nv_get_vgpu_type_id(NULL, NULL, &vgpu_type_id, mtype);
    if (status != NV_OK)
        return -ENXIO;

    status = rm_vgpu_vfio_ops.get_instances(pdev, vgpu_type_id, buf);
    if (status != NV_OK)
        return -EIO;

    ret = kstrtouint(buf, 10, &available_instances);
    if (ret)
        return ret;

    return available_instances;
}

static ssize_t nv_vfio_mdev_show_description(struct mdev_type *mtype, char *buf)
{
    NvU32 vgpu_type_id = 0;
    struct device *dev = NV_MTYPE_GET_PARENT_DEV(mtype);
    struct pci_dev *pdev = to_pci_dev(dev);
    NV_STATUS status;

    status = nv_get_vgpu_type_id(NULL, NULL, &vgpu_type_id, mtype);
    if (status != NV_OK)
        return -ENXIO;

    status = rm_vgpu_vfio_ops.get_description(pdev, vgpu_type_id, buf);
    if (status != NV_OK)
        return -EIO;

    return strlen(buf);
}
#endif

struct mdev_driver nv_vfio_mdev_driver = {
    .driver = {
        .name       = "nvidia-vgpu-vfio",
        .owner      = THIS_MODULE,
        .mod_name   = KBUILD_MODNAME,
        .dev_groups = vgpu_dev_groups,
    },
    .probe  = nv_vfio_mdev_probe,
    .remove = nv_vfio_mdev_remove,
#if defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    .device_api = VFIO_DEVICE_API_PCI_STRING,
    .get_available = nv_vfio_mdev_get_available,
    .show_description = nv_vfio_mdev_show_description,
#endif
};

NV_STATUS nv_vgpu_set_mdev_fops(struct phys_dev_s *phys_dev)
{

#if !defined(NV_MDEV_PARENT_STRUCT_PRESENT)

#if defined(NV_MDEV_DRIVER_HAS_SUPPORTED_TYPE_GROUPS)
    NV_KMALLOC(phys_dev->vgpu_fops, sizeof(struct mdev_driver));
#elif defined(NV_MDEV_PARENT_OPS_STRUCT_PRESENT)
    NV_KMALLOC(phys_dev->vgpu_fops, sizeof(struct mdev_parent_ops));
#else
    NV_KMALLOC(phys_dev->vgpu_fops, sizeof(struct parent_ops));
#endif

    if (phys_dev->vgpu_fops == NULL)
        return NV_ERR_NO_MEMORY;

#if defined(NV_MDEV_DRIVER_HAS_SUPPORTED_TYPE_GROUPS)
    memset(phys_dev->vgpu_fops, 0, sizeof(struct mdev_driver));
#elif defined(NV_MDEV_PARENT_OPS_STRUCT_PRESENT)
    memset(phys_dev->vgpu_fops, 0, sizeof(struct mdev_parent_ops));
#else
    memset(phys_dev->vgpu_fops, 0, sizeof(struct parent_ops));
#endif

#if defined(NV_MDEV_DRIVER_HAS_SUPPORTED_TYPE_GROUPS)
    memcpy((void *)phys_dev->vgpu_fops, &nv_vfio_mdev_driver, sizeof(struct mdev_driver));
#else
    phys_dev->vgpu_fops->device_driver = &nv_vfio_mdev_driver;
#endif

#if !defined(NV_MDEV_DRIVER_HAS_SUPPORTED_TYPE_GROUPS)
    /* .owner needs to be registered with mdev module till kernel v5.19 */
    phys_dev->vgpu_fops->owner                 = THIS_MODULE;
#endif

    /* .supported_type_groups needs to be registered with mdev module till kernel v6.1 */
    phys_dev->vgpu_fops->supported_type_groups = phys_dev->vgpu_type_groups;

#endif /* !NV_MDEV_PARENT_STRUCT_PRESENT */

    return NV_OK;
}

int nv_vfio_mdev_register_driver()
{
    return mdev_register_driver(&nv_vfio_mdev_driver);
}

void nv_vfio_mdev_unregister_driver()
{
    mdev_unregister_driver(&nv_vfio_mdev_driver);
}

#else /* NV_USE_MODERN_MDEV */

/* Old mdev codepath for kernel < v5.14 */

static vgpu_dev_t *nv_get_vgpu_from_mdev(struct mdev_device *mdev)
{
    vgpu_dev_t *vgpu_dev = NULL;

    if (!mdev)
    {
        NV_VGPU_LOG(VGPU_ERR, "No mdev device found\n");
        return NULL;
    }

    vgpu_dev = mdev_get_drvdata(mdev);
    if (!vgpu_dev)
        NV_VGPU_LOG(VGPU_ERR, "%s No vGPU device found\n", dev_name(NV_GET_MDEV_DEV(mdev)));

    return vgpu_dev;
}

static int nv_vfio_mdev_open(struct mdev_device *mdev)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    return nv_vgpu_vfio_open(vgpu_dev);
}

static void nv_vfio_mdev_close(struct mdev_device *mdev)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    nv_vgpu_vfio_close(vgpu_dev);
}

static int nv_vfio_mdev_mmap(struct mdev_device *mdev, struct vm_area_struct *vma)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    return nv_vgpu_vfio_mmap(vgpu_dev, vma);
}

static long nv_vfio_mdev_ioctl(struct mdev_device *mdev, unsigned int cmd,
                               unsigned long arg)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    return nv_vgpu_vfio_ioctl(vgpu_dev, cmd, arg);
}

static ssize_t nv_vfio_mdev_read(struct mdev_device *mdev, char __user *buf,
                                 size_t count, loff_t *ppos)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    return nv_vgpu_vfio_read(vgpu_dev, buf, count, ppos);
}

static ssize_t nv_vfio_mdev_write(struct mdev_device *mdev, const char __user *buf,
                                  size_t count, loff_t *ppos)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    return nv_vgpu_vfio_write(vgpu_dev, buf, count, ppos);
}

#if defined(NV_MDEV_GET_TYPE_GROUP_ID_PRESENT) || defined(NV_MDEV_PARENT_STRUCT_PRESENT)
static int nv_vfio_mdev_create(struct mdev_device *mdev)
#else
static int nv_vfio_mdev_create(struct kobject *kobj, struct mdev_device *mdev)
#endif
{
    vgpu_dev_t *vgpu_dev = NULL;
    int ret = 0;

    if (!mdev)
        return -EINVAL;

    ret = nv_alloc_vgpu_dev(&vgpu_dev);
    if (ret)
        return ret;

    nv_init_vgpu_dev(vgpu_dev, mdev);

#if defined(NV_MDEV_GET_TYPE_GROUP_ID_PRESENT) || defined(NV_MDEV_PARENT_STRUCT_PRESENT)
    ret = nv_vgpu_vfio_mdev_create(vgpu_dev);
#else
    ret = nv_vgpu_vfio_mdev_create(kobj, vgpu_dev);
#endif
    if (ret)
        nv_free_vgpu_dev(vgpu_dev);

    mdev_set_drvdata(vgpu_dev->mdev, vgpu_dev);
    return ret;
}

static int nv_vfio_mdev_destroy(struct mdev_device *mdev)
{
    vgpu_dev_t *vgpu_dev = nv_get_vgpu_from_mdev(mdev);
    int ret = 0;

    if (!vgpu_dev)
        return -EINVAL;

    ret = nv_vgpu_vfio_mdev_destroy(vgpu_dev);

    mdev_set_drvdata(mdev, NULL);
    nv_free_vgpu_dev(vgpu_dev);

    return ret;
}

NV_STATUS nv_vgpu_set_mdev_fops(struct phys_dev_s *phys_dev)
{
#if defined(NV_MDEV_PARENT_OPS_STRUCT_PRESENT)
    NV_KMALLOC(phys_dev->vgpu_fops, sizeof(struct mdev_parent_ops));
#else
    NV_KMALLOC(phys_dev->vgpu_fops, sizeof(struct parent_ops));
#endif
    if (phys_dev->vgpu_fops == NULL)
        return NV_ERR_NO_MEMORY;

#if defined(NV_MDEV_PARENT_OPS_STRUCT_PRESENT)
    memset(phys_dev->vgpu_fops, 0, sizeof(struct mdev_parent_ops));
#else
    memset(phys_dev->vgpu_fops, 0, sizeof(struct parent_ops));
#endif

    /* vgpu fops to be registed to mdev module */
    phys_dev->vgpu_fops->owner                 = THIS_MODULE;
    phys_dev->vgpu_fops->supported_type_groups = phys_dev->vgpu_type_groups;
    phys_dev->vgpu_fops->create                = nv_vfio_mdev_create;
    phys_dev->vgpu_fops->remove                = nv_vfio_mdev_destroy;
    phys_dev->vgpu_fops->mdev_attr_groups      = vgpu_dev_groups;
    phys_dev->vgpu_fops->read                  = nv_vfio_mdev_read;
    phys_dev->vgpu_fops->write                 = nv_vfio_mdev_write;
    phys_dev->vgpu_fops->ioctl                 = nv_vfio_mdev_ioctl;
    phys_dev->vgpu_fops->mmap                  = nv_vfio_mdev_mmap;

#if defined(NV_MDEV_PARENT_OPS_HAS_OPEN_DEVICE)
    phys_dev->vgpu_fops->open_device           = nv_vfio_mdev_open;
    phys_dev->vgpu_fops->close_device          = nv_vfio_mdev_close;
#else
    phys_dev->vgpu_fops->open                  = nv_vfio_mdev_open;
    phys_dev->vgpu_fops->release               = nv_vfio_mdev_close;
#endif

    return NV_OK;
}

#endif /* NV_USE_MODERN_MDEV */

#endif /* NV_VGPU_KVM_BUILD &&  NV_MDEV_PRESENT */
